"""This module provides classes representing grammars."""

import json
from typing import Any, Dict, List, Optional, Tuple, Type, Union, overload

from pydantic import BaseModel
from typing_extensions import deprecated

from .base import XGRObject, _core
from .structural_tag import StructuralTag, StructuralTagItem


def _convert_instance_to_str(instance: Union[str, Dict[str, Any], StructuralTag]) -> str:
    """Convert a instance to a string representation. It returns the schema in string format because
    it's faster to send to C++.

    This function handles different instance input types and converts them to a JSON string:
    - StructuralTag.
    - String inputs are returned as-is (assumed to be valid JSON)
    - Dictionary inputs are converted to JSON strings

    Parameters
    ----------
    instance : Union[str, StructuralTag, Dict[str, Any]]
        The instance to convert, which can be a StructuralTag,
        a JSON schema string, or a dictionary representing a JSON schema.

    Returns
    -------
    str
        The JSON schema as a string.

    Raises
    ------
    ValueError
        When the instance type is not supported.
    TypeError
        When he dictionary is not serializable.
    """
    if isinstance(instance, dict):
        return json.dumps(instance)
    elif isinstance(instance, str):
        return instance
    elif isinstance(instance, StructuralTag):
        return instance.model_dump_json()
    else:
        raise ValueError("Invalid instance type")


def _convert_schema_to_str(schema: Union[str, Type[BaseModel], Dict[str, Any]]) -> str:
    """Convert a schema to a string representation. It returns the schema in string format because
    it's faster to send to C++.

    This function handles different schema input types and converts them to a JSON string:
    - Pydantic models are converted using their schema methods
    - String inputs are returned as-is (assumed to be valid JSON)
    - Dictionary inputs are converted to JSON strings

    Parameters
    ----------
    schema : Union[str, Type[BaseModel], Dict[str, Any]]
        The schema to convert, which can be a Pydantic model class,
        a JSON schema string, or a dictionary representing a JSON schema.

    Returns
    -------
    str
        The JSON schema as a string.

    Raises
    ------
    ValueError
        When the schema type is not supported.
    TypeError
        When the dictionary is not serializable.
    """
    if isinstance(schema, type) and issubclass(schema, BaseModel):
        if hasattr(schema, "model_json_schema"):
            return json.dumps(schema.model_json_schema())
        if hasattr(schema, "schema_json"):
            return json.dumps(schema.schema_json())
        else:
            raise ValueError("The schema should have a model_json_schema or json_schema method.")
    elif isinstance(schema, str):
        return schema
    elif isinstance(schema, dict):
        return json.dumps(schema)
    else:
        raise ValueError("The schema should be a string or a Pydantic model.")


def _get_structural_tag_str_from_args(args: List[Any], kwargs: Dict[str, Any]) -> str:
    """Get the structural tag string from the arguments. It returns the structural tag in string
    format because it's faster to send to C++.

    Parameters
    ----------
    args : List[Any]
        The positional arguments.
    kwargs : Dict[str, Any]
        The keyword arguments.

    Returns
    -------
    str
        The structural tag string.

    Raises
    ------
    TypeError
        When the arguments are invalid.
    """
    if len(args) == 1:
        if isinstance(args[0], (str, dict, StructuralTag)):
            return _convert_instance_to_str(args[0])
        else:
            raise TypeError("Invalid argument type for from_structural_tag")
    elif len(args) == 2 and isinstance(args[0], list) and isinstance(args[1], list):
        return StructuralTag.from_legacy_structural_tag(args[0], args[1]).model_dump_json(
            indent=None
        )
    elif "structural_tag" in kwargs:
        return _convert_instance_to_str(kwargs["structural_tag"])
    elif "tags" in kwargs and "triggers" in kwargs:
        return StructuralTag.from_legacy_structural_tag(
            kwargs["tags"], kwargs["triggers"]
        ).model_dump_json(indent=None)
    else:
        raise TypeError("Invalid arguments for from_structural_tag")


class Grammar(XGRObject):
    """This class represents a grammar object in XGrammar, and can be used later in the
    grammar-guided generation.

    The Grammar object supports context-free grammar (CFG). EBNF (extended Backus-Naur Form) is
    used as the format of the grammar. There are many specifications for EBNF in the literature,
    and we follow the specification of GBNF (GGML BNF) in
    https://github.com/ggerganov/llama.cpp/blob/master/grammars/README.md.

    When printed, the grammar will be converted to GBNF format.
    """

    def __str__(self) -> str:
        """Print the BNF grammar to a string, in EBNF format.

        Returns
        -------
        grammar_string : str
            The BNF grammar string.
        """
        return self._handle.to_string()

    @staticmethod
    def from_ebnf(ebnf_string: str, *, root_rule_name: str = "root") -> "Grammar":
        """Construct a grammar from EBNF string. The EBNF string should follow the format
        in https://github.com/ggerganov/llama.cpp/blob/master/grammars/README.md.

        Parameters
        ----------
        ebnf_string : str
            The grammar string in EBNF format.

        root_rule_name : str, default: "root"
            The name of the root rule in the grammar.

        Raises
        ------
        RuntimeError
            When converting the regex pattern fails, with details about the parsing error.
        """
        return Grammar._create_from_handle(_core.Grammar.from_ebnf(ebnf_string, root_rule_name))

    @staticmethod
    def from_json_schema(
        schema: Union[str, Type[BaseModel], Dict[str, Any]],
        *,
        any_whitespace: bool = True,
        indent: Optional[int] = None,
        separators: Optional[Tuple[str, str]] = None,
        strict_mode: bool = True,
        max_whitespace_cnt: Optional[int] = None,
        print_converted_ebnf: bool = False,
    ) -> "Grammar":
        """Construct a grammar from JSON schema. Pydantic model or JSON schema string can be
        used to specify the schema.

        It allows any whitespace by default. If user want to specify the format of the JSON,
        set `any_whitespace` to False and use the `indent` and `separators` parameters. The
        meaning and the default values of the parameters follows the convention in json.dumps().

        It internally converts the JSON schema to a EBNF grammar.

        Parameters
        ----------
        schema : Union[str, Type[BaseModel], Dict[str, Any]]
            The schema string or Pydantic model or JSON schema dict.

        any_whitespace : bool, default: True
            Whether to use any whitespace. If True, the generated grammar will ignore the
            indent and separators parameters, and allow any whitespace.

        indent : Optional[int], default: None
            The number of spaces for indentation. If None, the output will be in one line.

            Note that specifying the indentation means forcing the LLM to generate JSON strings
            strictly formatted. However, some models may tend to generate JSON strings that
            are not strictly formatted. In this case, forcing the LLM to generate strictly
            formatted JSON strings may degrade the generation quality. See
            <https://github.com/sgl-project/sglang/issues/2216#issuecomment-2516192009> for more
            details.

        separators : Optional[Tuple[str, str]], default: None
            Two separators used in the schema: comma and colon. Examples: (",", ":"), (", ", ": ").
            If None, the default separators will be used: (",", ": ") when the indent is not None,
            and (", ", ": ") otherwise.

        strict_mode : bool, default: True
            Whether to use strict mode. In strict mode, the generated grammar will not allow
            properties and items that is not specified in the schema. This is equivalent to
            setting unevaluatedProperties and unevaluatedItems to false.

            This helps LLM to generate accurate output in the grammar-guided generation with JSON
            schema.

        max_whitespace_cnt : Optional[int], default: None
            The maximum number of whitespace characters allowed between elements, such like keys, values, separators and so on.
            If None, there is no limit on the number of whitespace characters.
            If specified, it will limit the number of whitespace characters to at most max_whitespace_cnt.
            It should be a positive integer.

        print_converted_ebnf : bool, default: False
            If True, the converted EBNF string will be printed. For debugging purposes.

        Returns
        -------
        grammar : Grammar
            The constructed grammar.

        Raises
        ------
        RuntimeError
            When converting the json schema fails, with details about the parsing error.
        """
        schema_str = _convert_schema_to_str(schema)
        return Grammar._create_from_handle(
            _core.Grammar.from_json_schema(
                schema_str,
                any_whitespace,
                indent,
                separators,
                strict_mode,
                max_whitespace_cnt,
                print_converted_ebnf,
            )
        )

    @staticmethod
    def from_regex(regex_string: str, *, print_converted_ebnf: bool = False) -> "Grammar":
        """Create a grammar from a regular expression string.

        Parameters
        ----------
        regex_string : str
            The regular expression pattern to create the grammar from.

        print_converted_ebnf : bool, default: False
            This method will convert the regex pattern to EBNF first. If this is true, the converted
            EBNF string will be printed. For debugging purposes. Default: False.

        Returns
        -------
        grammar : Grammar
            The constructed grammar from the regex pattern.

        Raises
        ------
        RuntimeError
            When parsing the regex pattern fails, with details about the parsing error.
        """
        return Grammar._create_from_handle(
            _core.Grammar.from_regex(regex_string, print_converted_ebnf)
        )

    @overload
    @staticmethod
    def from_structural_tag(
        structural_tag: Union[StructuralTag, str, Dict[str, Any]]
    ) -> "Grammar": ...

    @overload
    @staticmethod
    @deprecated(
        "from_structural_tag(tags, triggers) is deprecated. Construct structural tag with the "
        "StructuralTag class instead."
    )
    def from_structural_tag(tags: List[StructuralTagItem], triggers: List[str]) -> "Grammar": ...

    @staticmethod
    def from_structural_tag(*args, **kwargs) -> "Grammar":
        """Create a grammar from a structural tag. See the Structural Tag Usage in XGrammar
        documentation for its usage.

        This method supports two calling patterns:

        1. Single structural tag parameter:
           from_structural_tag(structural_tag)

        2. Legacy pattern (deprecated):
           from_structural_tag(tags, triggers)

        Parameters
        ----------
        structural_tag : Union[StructuralTag, str, Dict[str, Any]]
            The structural tag either as a StructuralTag object, or a JSON string or a dictionary.

        tags : List[StructuralTagItem]
            (Deprecated) The structural tags. Use StructuralTag class instead.

        triggers : List[str]
            (Deprecated) The triggers. Use StructuralTag class instead.

        Returns
        -------
        grammar : Grammar
            The constructed grammar from the structural tag.

        Raises
        ------
        InvalidJSONError
            When the structural tag is not a valid JSON string.
        InvalidStructuralTagError
            When the structural tag is not valid.
        TypeError
            When the arguments are invalid.

        Notes
        -----
        The legacy pattern from_structural_tag(tags, triggers) is deprecated. Use the StructuralTag
        class to construct structural tags instead.

        For the deprecated pattern: The structural tag handles the dispatching of different grammars
        based on the tags and triggers: it initially allows any output, until a trigger is
        encountered, then dispatch to the corresponding tag; when the end tag is encountered, the
        grammar will allow any following output, until the next trigger is encountered. See the
        Advanced Topics of the Structural Tag in XGrammar documentation for its semantic.
        Structural Tag in XGrammar documentation for its semantic.
        """
        structural_tag_str = _get_structural_tag_str_from_args(args, kwargs)
        return Grammar._create_from_handle(_core.Grammar.from_structural_tag(structural_tag_str))

    @staticmethod
    def builtin_json_grammar() -> "Grammar":
        """Get the grammar of standard JSON. This is compatible with the official JSON grammar
        specification in https://www.json.org/json-en.html.

        Returns
        -------
        grammar : Grammar
            The JSON grammar.
        """
        return Grammar._create_from_handle(_core.Grammar.builtin_json_grammar())

    @staticmethod
    def concat(*grammars: "Grammar") -> "Grammar":
        """Create a grammar that matches the concatenation of the grammars in the list. That is
        equivalent to using the `+` operator to concatenate the grammars in the list.

        Parameters
        ----------
        grammars : List[Grammar]
            The grammars to create the concatenation of.

        Returns
        -------
        grammar : Grammar
            The concatenation of the grammars.
        """
        grammar_handles = [grammar._handle for grammar in grammars]
        return Grammar._create_from_handle(_core.Grammar.concat(grammar_handles))

    @staticmethod
    def union(*grammars: "Grammar") -> "Grammar":
        """Create a grammar that matches any of the grammars in the list. That is equivalent to
        using the `|` operator to concatenate the grammars in the list.

        Parameters
        ----------
        grammars : List[Grammar]
            The grammars to create the union of.

        Returns
        -------
        grammar : Grammar
            The union of the grammars.
        """
        grammar_handles = [grammar._handle for grammar in grammars]
        return Grammar._create_from_handle(_core.Grammar.union(grammar_handles))

    def serialize_json(self) -> str:
        """Serialize the grammar to a JSON string.

        Returns
        -------
        json_string : str
            The JSON string.
        """
        return self._handle.serialize_json()

    @staticmethod
    def deserialize_json(json_string: str) -> "Grammar":
        """Deserialize a grammar from a JSON string.

        Parameters
        ----------
        json_string : str
            The JSON string.

        Returns
        -------
        grammar : Grammar
            The deserialized grammar.

        Raises
        ------
        InvalidJSONError
            When the JSON string is invalid.
        DeserializeFormatError
            When the JSON string does not follow the serialization format of the grammar.
        DeserializeVersionError
            When the __VERSION__ field in the JSON string is not the same as the current version.
        """
        return Grammar._create_from_handle(_core.Grammar.deserialize_json(json_string))



"""This test uses the optimized JSON grammar provided by the grammar library."""

import sys

import pytest

import xgrammar as xgr


def test_grammar_union():
    grammar1 = xgr.Grammar.from_ebnf(
        """root ::= r1 | r2
r1 ::= "true" | ""
r2 ::= "false" | ""
"""
    )

    grammar2 = xgr.Grammar.from_ebnf(
        """root ::= "abc" | r1
r1 ::= "true" | r1
"""
    )

    grammar3 = xgr.Grammar.from_ebnf(
        """root ::= r1 | r2 | r3
r1 ::= "true" | r3
r2 ::= "false" | r3
r3 ::= "abc" | ""
"""
    )

    expected = """root ::= ((root_1) | (root_2) | (root_3))
root_1 ::= ((r1) | (r2))
r1 ::= ("" | ("true"))
r2 ::= ("" | ("false"))
root_2 ::= (("abc") | (r1_1))
r1_1 ::= (("true") | (r1_1))
root_3 ::= ((r1_2) | (r2_1) | (r3))
r1_2 ::= (("true") | (r3))
r2_1 ::= (("false") | (r3))
r3 ::= ("" | ("abc"))
"""

    union_grammar = xgr.Grammar.union(grammar1, grammar2, grammar3)
    assert str(union_grammar) == expected


def test_grammar_concat():
    grammar1 = xgr.Grammar.from_ebnf(
        """root ::= r1 | r2
r1 ::= "true" | ""
r2 ::= "false" | ""
"""
    )

    grammar2 = xgr.Grammar.from_ebnf(
        """root ::= "abc" | r1
r1 ::= "true" | r1
"""
    )

    grammar3 = xgr.Grammar.from_ebnf(
        """root ::= r1 | r2 | r3
r1 ::= "true" | r3
r2 ::= "false" | r3
r3 ::= "abc" | ""
"""
    )

    expected = """root ::= ((root_1 root_2 root_3))
root_1 ::= ((r1) | (r2))
r1 ::= ("" | ("true"))
r2 ::= ("" | ("false"))
root_2 ::= (("abc") | (r1_1))
r1_1 ::= (("true") | (r1_1))
root_3 ::= ((r1_2) | (r2_1) | (r3))
r1_2 ::= (("true") | (r3))
r2_1 ::= (("false") | (r3))
r3 ::= ("" | ("abc"))
"""

    concat_grammar = xgr.Grammar.concat(grammar1, grammar2, grammar3)
    assert str(concat_grammar) == expected


def test_grammar_union_with_stag():
    expected_grammar_union = r"""root ::= ((root_1_1) | (root_2))
basic_escape ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9]))
basic_string_sub ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub) | ("\\" basic_escape basic_string_sub)) (=([ \n\t]* [,}\]:]))
basic_any ::= ((basic_number) | (basic_string) | (basic_boolean) | (basic_null) | (basic_array) | (basic_object))
basic_integer ::= (("0") | (basic_integer_1 [1-9] [0-9]*))
basic_number ::= ((basic_number_1 basic_number_7 basic_number_3 basic_number_6))
basic_string ::= (("\"" basic_string_sub))
basic_boolean ::= (("true") | ("false"))
basic_null ::= (("null"))
basic_array ::= (("[" [ \n\t]* basic_any basic_array_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
basic_object ::= (("{" [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1 [ \n\t]* "}") | ("{" [ \n\t]* "}"))
root_1 ::= (("{" [ \n\t]* "\"arg\"" [ \n\t]* ":" [ \n\t]* basic_string [ \n\t]* "}") | ("{" [ \n\t]* "}"))
basic_integer_1 ::= ("" | ("-"))
basic_number_1 ::= ("" | ("-"))
basic_number_2 ::= (([0-9] basic_number_2) | ([0-9]))
basic_number_3 ::= ("" | ("." basic_number_2))
basic_number_4 ::= ("" | ([+\-]))
basic_number_5 ::= (([0-9] basic_number_5) | ([0-9]))
basic_number_6 ::= ("" | ([eE] basic_number_4 basic_number_5))
basic_array_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_any basic_array_1))
basic_object_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1))
basic_number_7 ::= (("0") | ([1-9] [0-9]*))
triggered_tags_group ::= (("" root_1 "end"))
triggered_tags ::= TagDispatch(
  ("start", triggered_tags_group),
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=true
)
root_1_1 ::= ((triggered_tags))
root_2 ::= (([a-z] root_2) | ([a-z]))
"""

    expected_grammar_concat = r"""root ::= ((root_1_1 root_2))
basic_escape ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9]))
basic_string_sub ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub) | ("\\" basic_escape basic_string_sub)) (=([ \n\t]* [,}\]:]))
basic_any ::= ((basic_number) | (basic_string) | (basic_boolean) | (basic_null) | (basic_array) | (basic_object))
basic_integer ::= (("0") | (basic_integer_1 [1-9] [0-9]*))
basic_number ::= ((basic_number_1 basic_number_7 basic_number_3 basic_number_6))
basic_string ::= (("\"" basic_string_sub))
basic_boolean ::= (("true") | ("false"))
basic_null ::= (("null"))
basic_array ::= (("[" [ \n\t]* basic_any basic_array_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
basic_object ::= (("{" [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1 [ \n\t]* "}") | ("{" [ \n\t]* "}"))
root_1 ::= (("{" [ \n\t]* "\"arg\"" [ \n\t]* ":" [ \n\t]* basic_string [ \n\t]* "}") | ("{" [ \n\t]* "}"))
basic_integer_1 ::= ("" | ("-"))
basic_number_1 ::= ("" | ("-"))
basic_number_2 ::= (([0-9] basic_number_2) | ([0-9]))
basic_number_3 ::= ("" | ("." basic_number_2))
basic_number_4 ::= ("" | ([+\-]))
basic_number_5 ::= (([0-9] basic_number_5) | ([0-9]))
basic_number_6 ::= ("" | ([eE] basic_number_4 basic_number_5))
basic_array_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_any basic_array_1))
basic_object_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1))
basic_number_7 ::= (("0") | ([1-9] [0-9]*))
triggered_tags_group ::= (("" root_1 "end"))
triggered_tags ::= TagDispatch(
  ("start", triggered_tags_group),
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=true
)
root_1_1 ::= ((triggered_tags))
root_2 ::= (([a-z] root_2) | ([a-z]))
"""
    start = "start"
    schema = {"type": "object", "properties": {"arg": {"type": "string"}}}
    end = "end"
    tag = xgr.StructuralTagItem(begin=start, schema=schema, end=end)
    triggers = [start]
    stag_grammar = xgr.Grammar.from_structural_tag([tag], triggers)
    start_grammar = xgr.Grammar.from_ebnf("root ::= [a-z] root | [a-z]")
    grammar_union = xgr.Grammar.union(stag_grammar, start_grammar)
    assert str(grammar_union) == expected_grammar_union
    grammar_concat = xgr.Grammar.concat(stag_grammar, start_grammar)
    assert str(grammar_concat) == expected_grammar_concat


if __name__ == "__main__":
    pytest.main(sys.argv)



import json
import sys
import threading
import time
from typing import List

import pytest
from pydantic import BaseModel
from transformers import AutoTokenizer

import xgrammar as xgr
from xgrammar.testing import _get_masked_tokens_from_bitmask, _is_grammar_accept_string


def test_utf8():
    # Test utf8-encoded string with structural tags
    class Schema(BaseModel):
        arg1: str
        arg2: int

    tags = [
        xgr.StructuralTagItem(begin="，，", schema=Schema, end="。"),
        xgr.StructuralTagItem(begin="，！", schema=Schema, end="。。"),
        xgr.StructuralTagItem(begin="，，？", schema=Schema, end="。。。"),
        xgr.StructuralTagItem(begin="｜｜？", schema=Schema, end="｜？｜"),
    ]
    triggers = ["，", "｜｜"]

    grammar = xgr.Grammar.from_structural_tag(tags, triggers)

    accepted_inputs = [
        '这是无用的内容，，{"arg1": "你好，世界！", "arg2": 0}。这是无用的内容',
        '这是无用的内容，！{"arg1": "こんにちは！", "arg2": 1}。。这是无用的内容',
        '这是无用的内容，，？{"arg1": "안녕하세요！", "arg2": 2}。。。这是无用的内容，！{"arg1": "안녕하세요！", "arg2": 3}。。',
        '这是无用的内容｜｜？{"arg1": "။စ်န, ်ပြ！", "arg2": 0}｜？｜｜｜？{"arg1": "။စ်န, ်ပြ", "arg2": 0}｜？｜',
    ]
    for input_str in accepted_inputs:
        assert _is_grammar_accept_string(grammar, input_str, print_time=True)


expected_grammar_test_structural_tag_after_optimization = r"""basic_escape ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9])) (=(basic_string_sub))
basic_string_sub ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub) | ("\\" basic_escape basic_string_sub)) (=([ \n\t]* [,}\]:]))
basic_integer ::= (("0") | (basic_integer_1 [1-9] [0-9]*))
basic_string ::= (("\"" basic_string_sub)) (=(root_part_0 [ \n\t]* "}"))
root_part_0 ::= (([ \n\t]* "," [ \n\t]* "\"arg2\"" [ \n\t]* ":" [ \n\t]* basic_integer)) (=([ \n\t]* "}"))
root ::= (("{" [ \n\t]* "\"arg1\"" [ \n\t]* ":" [ \n\t]* basic_string root_part_0 [ \n\t]* "}")) (=("</function>"))
basic_integer_1 ::= ("" | ("-")) (=([1-9] [0-9]*))
basic_escape_1 ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9])) (=(basic_string_sub_1))
basic_string_sub_1 ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub_1) | ("\\" basic_escape_1 basic_string_sub_1)) (=([ \n\t]* [,}\]:]))
basic_integer_2 ::= (("0") | (basic_integer_1_1 [1-9] [0-9]*))
basic_string_1 ::= (("\"" basic_string_sub_1)) (=(root_part_0_1 [ \n\t]* "}"))
root_part_0_1 ::= (([ \n\t]* "," [ \n\t]* "\"arg2\"" [ \n\t]* ":" [ \n\t]* basic_integer_2)) (=([ \n\t]* "}"))
root_1 ::= (("{" [ \n\t]* "\"arg1\"" [ \n\t]* ":" [ \n\t]* basic_string_1 root_part_0_1 [ \n\t]* "}")) (=("</function>"))
basic_integer_1_1 ::= ("" | ("-")) (=([1-9] [0-9]*))
basic_escape_2 ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9])) (=(basic_string_sub_2))
basic_string_sub_2 ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub_2) | ("\\" basic_escape_2 basic_string_sub_2)) (=([ \n\t]* [,}\]:]))
basic_number_9 ::= ((basic_number_1_2 basic_number_7_2 basic_number_3_2 basic_number_6_2)) (=(root_part_0_2 [ \n\t]* "}"))
basic_string_2 ::= (("\"" basic_string_sub_2))
root_prop_1 ::= (("[" [ \n\t]* basic_string_2 root_prop_1_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
root_part_0_2 ::= (([ \n\t]* "," [ \n\t]* "\"arg4\"" [ \n\t]* ":" [ \n\t]* root_prop_1)) (=([ \n\t]* "}"))
root_2 ::= (("{" [ \n\t]* "\"arg3\"" [ \n\t]* ":" [ \n\t]* basic_number_9 root_part_0_2 [ \n\t]* "}")) (=("</function>"))
basic_number_1_2 ::= ("" | ("-")) (=(basic_number_7_2 basic_number_3_2 basic_number_6_2))
basic_number_2_2 ::= (([0-9] basic_number_2_2) | ([0-9]))
basic_number_3_2 ::= ("" | ("." basic_number_2_2)) (=(basic_number_6_2))
basic_number_4_2 ::= ("" | ([+\-])) (=(basic_number_5_2))
basic_number_5_2 ::= (([0-9] basic_number_5_2) | ([0-9]))
basic_number_6_2 ::= ("" | ([eE] basic_number_4_2 basic_number_5_2))
root_prop_1_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string_2 root_prop_1_1)) (=([ \n\t]* "]"))
basic_number_7_2 ::= (("0") | ([1-9] [0-9]*)) (=(basic_number_3_2 basic_number_6_2))
triggered_tags_group ::= (("1>" root "</function>") | ("2>" root_1 "</function>"))
triggered_tags_group_1 ::= ((">" root_2 "</function>"))
triggered_tags ::= TagDispatch(
  ("<function=f", triggered_tags_group),
  ("<function=g", triggered_tags_group_1),
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=true
)
root_3 ::= ((triggered_tags))
"""

expected_grammar_test_structural_tag_before_optimization = r"""basic_escape ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9]))
basic_string_sub ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub) | ("\\" basic_escape basic_string_sub)) (=([ \n\t]* [,}\]:]))
basic_any ::= ((basic_number) | (basic_string) | (basic_boolean) | (basic_null) | (basic_array) | (basic_object))
basic_integer ::= (("0") | (basic_integer_1 [1-9] [0-9]*))
basic_number ::= ((basic_number_1 basic_number_7 basic_number_3 basic_number_6))
basic_string ::= (("\"" basic_string_sub))
basic_boolean ::= (("true") | ("false"))
basic_null ::= (("null"))
basic_array ::= (("[" [ \n\t]* basic_any basic_array_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
basic_object ::= (("{" [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1 [ \n\t]* "}") | ("{" [ \n\t]* "}"))
root_part_0 ::= (([ \n\t]* "," [ \n\t]* "\"arg2\"" [ \n\t]* ":" [ \n\t]* basic_integer))
root ::= (("{" [ \n\t]* "\"arg1\"" [ \n\t]* ":" [ \n\t]* basic_string root_part_0 [ \n\t]* "}"))
basic_integer_1 ::= ("" | ("-"))
basic_number_1 ::= ("" | ("-"))
basic_number_2 ::= (([0-9] basic_number_2) | ([0-9]))
basic_number_3 ::= ("" | ("." basic_number_2))
basic_number_4 ::= ("" | ([+\-]))
basic_number_5 ::= (([0-9] basic_number_5) | ([0-9]))
basic_number_6 ::= ("" | ([eE] basic_number_4 basic_number_5))
basic_array_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_any basic_array_1))
basic_object_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string [ \n\t]* ":" [ \n\t]* basic_any basic_object_1))
basic_number_7 ::= (("0") | ([1-9] [0-9]*))
basic_escape_1 ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9]))
basic_string_sub_1 ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub_1) | ("\\" basic_escape_1 basic_string_sub_1)) (=([ \n\t]* [,}\]:]))
basic_any_1 ::= ((basic_number_8) | (basic_string_1) | (basic_boolean_1) | (basic_null_1) | (basic_array_2) | (basic_object_2))
basic_integer_2 ::= (("0") | (basic_integer_1_1 [1-9] [0-9]*))
basic_number_8 ::= ((basic_number_1_1 basic_number_7_1 basic_number_3_1 basic_number_6_1))
basic_string_1 ::= (("\"" basic_string_sub_1))
basic_boolean_1 ::= (("true") | ("false"))
basic_null_1 ::= (("null"))
basic_array_2 ::= (("[" [ \n\t]* basic_any_1 basic_array_1_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
basic_object_2 ::= (("{" [ \n\t]* basic_string_1 [ \n\t]* ":" [ \n\t]* basic_any_1 basic_object_1_1 [ \n\t]* "}") | ("{" [ \n\t]* "}"))
root_part_0_1 ::= (([ \n\t]* "," [ \n\t]* "\"arg2\"" [ \n\t]* ":" [ \n\t]* basic_integer_2))
root_1 ::= (("{" [ \n\t]* "\"arg1\"" [ \n\t]* ":" [ \n\t]* basic_string_1 root_part_0_1 [ \n\t]* "}"))
basic_integer_1_1 ::= ("" | ("-"))
basic_number_1_1 ::= ("" | ("-"))
basic_number_2_1 ::= (([0-9] basic_number_2_1) | ([0-9]))
basic_number_3_1 ::= ("" | ("." basic_number_2_1))
basic_number_4_1 ::= ("" | ([+\-]))
basic_number_5_1 ::= (([0-9] basic_number_5_1) | ([0-9]))
basic_number_6_1 ::= ("" | ([eE] basic_number_4_1 basic_number_5_1))
basic_array_1_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_any_1 basic_array_1_1))
basic_object_1_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string_1 [ \n\t]* ":" [ \n\t]* basic_any_1 basic_object_1_1))
basic_number_7_1 ::= (("0") | ([1-9] [0-9]*))
basic_escape_2 ::= (([\"\\/bfnrt]) | ("u" [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9] [A-Fa-f0-9]))
basic_string_sub_2 ::= (("\"") | ([^\0-\x1f\"\\\r\n] basic_string_sub_2) | ("\\" basic_escape_2 basic_string_sub_2)) (=([ \n\t]* [,}\]:]))
basic_any_2 ::= ((basic_number_9) | (basic_string_2) | (basic_boolean_2) | (basic_null_2) | (basic_array_3) | (basic_object_3))
basic_integer_3 ::= (("0") | (basic_integer_1_2 [1-9] [0-9]*))
basic_number_9 ::= ((basic_number_1_2 basic_number_7_2 basic_number_3_2 basic_number_6_2))
basic_string_2 ::= (("\"" basic_string_sub_2))
basic_boolean_2 ::= (("true") | ("false"))
basic_null_2 ::= (("null"))
basic_array_3 ::= (("[" [ \n\t]* basic_any_2 basic_array_1_2 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
basic_object_3 ::= (("{" [ \n\t]* basic_string_2 [ \n\t]* ":" [ \n\t]* basic_any_2 basic_object_1_2 [ \n\t]* "}") | ("{" [ \n\t]* "}"))
root_prop_1 ::= (("[" [ \n\t]* basic_string_2 root_prop_1_1 [ \n\t]* "]") | ("[" [ \n\t]* "]"))
root_part_0_2 ::= (([ \n\t]* "," [ \n\t]* "\"arg4\"" [ \n\t]* ":" [ \n\t]* root_prop_1))
root_2 ::= (("{" [ \n\t]* "\"arg3\"" [ \n\t]* ":" [ \n\t]* basic_number_9 root_part_0_2 [ \n\t]* "}"))
basic_integer_1_2 ::= ("" | ("-"))
basic_number_1_2 ::= ("" | ("-"))
basic_number_2_2 ::= (([0-9] basic_number_2_2) | ([0-9]))
basic_number_3_2 ::= ("" | ("." basic_number_2_2))
basic_number_4_2 ::= ("" | ([+\-]))
basic_number_5_2 ::= (([0-9] basic_number_5_2) | ([0-9]))
basic_number_6_2 ::= ("" | ([eE] basic_number_4_2 basic_number_5_2))
basic_array_1_2 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_any_2 basic_array_1_2))
basic_object_1_2 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string_2 [ \n\t]* ":" [ \n\t]* basic_any_2 basic_object_1_2))
root_prop_1_1 ::= ("" | ([ \n\t]* "," [ \n\t]* basic_string_2 root_prop_1_1))
basic_number_7_2 ::= (("0") | ([1-9] [0-9]*))
triggered_tags_group ::= (("1>" root "</function>") | ("2>" root_1 "</function>"))
triggered_tags_group_1 ::= ((">" root_2 "</function>"))
triggered_tags ::= TagDispatch(
  ("<function=f", triggered_tags_group),
  ("<function=g", triggered_tags_group_1),
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=true
)
root_3 ::= ((triggered_tags))
"""


def test_structural_tag():
    class Schema1(BaseModel):
        arg1: str
        arg2: int

    class Schema2(BaseModel):
        arg3: float
        arg4: List[str]

    tags = [
        xgr.StructuralTagItem(begin="<function=f1>", schema=Schema1, end="</function>"),
        xgr.StructuralTagItem(begin="<function=f2>", schema=Schema1, end="</function>"),
        xgr.StructuralTagItem(begin="<function=g>", schema=Schema2, end="</function>"),
    ]
    # in real cases, we should use one trigger: "<function=" and dispatch to two tags
    # but here we use two triggers for testing such cases
    triggers = ["<function=f", "<function=g"]

    grammar = xgr.Grammar.from_structural_tag(tags, triggers)
    assert str(grammar) == expected_grammar_test_structural_tag_before_optimization

    accepted_inputs = [
        '<function=f1>{"arg1": "abc", "arg2": 1}</function>',
        '<function=g>{"arg3": 1.23, "arg4": ["a", "b", "c"]}</function>',
        '<function=f2>{"arg1": "abc", "arg2": 1}</function><function=g>{"arg3": 1.23, "arg4": ["a", "b", "c"]}</function>',
        'hhhh<function=g>{"arg3": 1.23, "arg4": ["a", "b", "c"]}</function>haha<function=f1>{"arg1": "abc", "arg2": 1}</function>123',
    ]
    for input in accepted_inputs:
        assert _is_grammar_accept_string(grammar, input, print_time=True)


def test_structural_tag_compiler():
    class Schema1(BaseModel):
        arg1: str
        arg2: int

    class Schema2(BaseModel):
        arg3: float
        arg4: List[str]

    tags = [
        xgr.StructuralTagItem(begin="<function=f1>", schema=Schema1, end="</function>"),
        xgr.StructuralTagItem(begin="<function=f2>", schema=Schema1, end="</function>"),
        xgr.StructuralTagItem(begin="<function=g>", schema=Schema2, end="</function>"),
    ]

    # in real cases, we should use one trigger: "<function=" and dispatch to two tags
    # but here we use two triggers for testing such cases
    triggers = ["<function=f", "<function=g"]

    compiler = xgr.GrammarCompiler(xgr.TokenizerInfo([]))
    compiled_grammar = compiler.compile_structural_tag(tags, triggers)
    assert str(compiled_grammar.grammar) == expected_grammar_test_structural_tag_after_optimization


@pytest.mark.hf_token_required
def test_structural_tag_mask_gen():
    # Define schemas for the test
    class Schema1(BaseModel):
        arg1: str
        arg2: int

    class Schema2(BaseModel):
        arg3: float
        arg4: List[str]

    # Set up grammar from schemas
    tags = [
        xgr.StructuralTagItem(
            begin="<function=f>", schema=json.dumps(Schema1.model_json_schema()), end="</function>"
        ),
        xgr.StructuralTagItem(
            begin="<function=g>", schema=json.dumps(Schema2.model_json_schema()), end="</function>"
        ),
    ]
    triggers = ["<function=f", "<function=g"]

    # Set up tokenizer
    tokenizer_id = "meta-llama/Llama-3.1-8B-Instruct"
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_id, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)

    # Compile grammar and create matcher
    compiler = xgr.GrammarCompiler(tokenizer_info)
    time_start = time.monotonic_ns()
    compiled_grammar = compiler.compile_structural_tag(tags, triggers)
    matcher = xgr.GrammarMatcher(compiled_grammar)
    time_end = time.monotonic_ns()
    print(f"Time to compile grammar and init GrammarMatcher: {(time_end - time_start) / 1e3} us")

    # Test input string
    accepted_input = (
        'hhhh<function=g>{"arg3": 1.23, "arg4": ["a", "b", "c"]}</function>'
        'haha<function=f>{"arg1": "abc", "arg2": 1}</function>123'
    )
    dont_apply_mask_indices = [
        # fmt: off
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76,
        77, 78, 119, 120, 121, 122
        # fmt: on
    ]
    input_bytes = accepted_input.encode("utf-8")

    # Set up token bitmask for validation
    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    # Process input character by character
    for i, c in enumerate(input_bytes):
        # 1. Test token bitmask generation
        time_start = time.monotonic_ns()
        need_apply = matcher.fill_next_token_bitmask(token_bitmask)
        time_end = time.monotonic_ns()
        print(f"Time to fill_next_token_bitmask: {(time_end - time_start) / 1e3} us")
        assert need_apply == (i not in dont_apply_mask_indices)

        # 2. Verify token bitmask correctness
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            token_bitmask, tokenizer_info.vocab_size
        )
        # This checking does not support non-ascii characters for now
        token_id_for_next_char = tokenizer.convert_tokens_to_ids(chr(c))
        assert token_id_for_next_char not in rejected_token_ids

        # 3. Test character acceptance
        print("Accepting char:", bytes([c]))
        time_start = time.monotonic_ns()
        assert matcher.accept_string(bytes([c]))
        time_end = time.monotonic_ns()
        print(f"Time to accept_token: {(time_end - time_start) / 1e3} us")

    # Final verification - check that EOS token is allowed
    time_start = time.monotonic_ns()
    need_apply = matcher.fill_next_token_bitmask(token_bitmask)
    time_end = time.monotonic_ns()
    assert need_apply == (len(input_bytes) not in dont_apply_mask_indices)
    print(f"Time to fill_next_token_bitmask: {(time_end - time_start) / 1e3} us")
    rejected_token_ids = _get_masked_tokens_from_bitmask(token_bitmask, tokenizer_info.vocab_size)
    assert tokenizer.eos_token_id not in rejected_token_ids


def test_empty_tag_dispatch():
    grammar_str = """root ::= TagDispatch(
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=true
)
"""
    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "any string")
    assert _is_grammar_accept_string(grammar, "")
    assert _is_grammar_accept_string(grammar, "好")

    grammar_with_stop_str_str = """root ::= TagDispatch(
  stop_eos=false,
  stop_str=("end"),
  loop_after_dispatch=true
)
"""

    grammar_with_stop_str = xgr.Grammar.from_ebnf(grammar_with_stop_str_str)

    assert _is_grammar_accept_string(grammar_with_stop_str, "any stringend")
    assert _is_grammar_accept_string(grammar_with_stop_str, "end")
    assert _is_grammar_accept_string(grammar_with_stop_str, "好end")

    assert not _is_grammar_accept_string(grammar_with_stop_str, "aaa")


@pytest.mark.hf_token_required
def test_utf8_structural_tag_begin_end():
    model = "deepseek-ai/DeepSeek-V3-0324"
    tokenizer = AutoTokenizer.from_pretrained(model)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    compiler = xgr.GrammarCompiler(tokenizer_info)
    structures = [
        xgr.StructuralTagItem(begin="<｜tool▁calls▁begin｜>", schema={}, end="<｜tool▁calls▁end｜>")
    ]
    triggers = ["<｜tool▁calls▁begin｜>"]
    _ = compiler.compile_structural_tag(structures, triggers)


@pytest.mark.hf_token_required
def test_pressure_structural_tag():
    model = "meta-llama/Llama-3.1-8B-Instruct"
    tokenizer = AutoTokenizer.from_pretrained(model, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    compiler = xgr.GrammarCompiler(tokenizer_info, max_threads=1)
    threads = []
    start = "start"
    schema = {"type": "object", "properties": {"arg": {"type": "string"}}}
    end = "end"

    def worker(idx: int):
        tag = xgr.StructuralTagItem(begin=start, schema=schema, end=end)
        triggers = [start]
        stag_grammar = xgr.Grammar.from_structural_tag([tag], triggers)
        start_grammar = xgr.Grammar.from_ebnf("root ::= [a-z] root | [a-z]")
        grammar = start_grammar
        for _ in range(idx):
            grammar = grammar.concat(grammar, start_grammar)
        final_grammar = xgr.Grammar.concat(grammar, stag_grammar)
        _ = compiler.compile_grammar(final_grammar)

    for i in range(128):
        t = threading.Thread(target=worker, args=(i,))
        threads.append(t)
        t.start()

    for t in threads:
        t.join()


if __name__ == "__main__":
    pytest.main(sys.argv)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from itertools import product
from typing import Any

import torch
from tabulate import tabulate
from tqdm import tqdm
from triton.testing import do_bench

from xgrammar.kernels.apply_token_bitmask_inplace_cuda import apply_token_bitmask_inplace_cuda
from xgrammar.kernels.apply_token_bitmask_inplace_torch_compile import (
    apply_token_bitmask_inplace_torch_compile,
)
from xgrammar.kernels.apply_token_bitmask_inplace_triton import apply_token_bitmask_inplace_triton
from xgrammar.testing import _bool_mask_to_bitmask

IMPL_TORCH_COMPILE: str = "Torch Compile"
IMPL_TRITON: str = "Triton"
IMPL_CUDA: str = "CUDA"

ALL_IMPLS: list[str] = [IMPL_TORCH_COMPILE, IMPL_TRITON, IMPL_CUDA]


def bench_single_impl(
    impl: str,
    logits: torch.Tensor,
    bitmask: torch.Tensor,
    logits_expected: torch.Tensor,
    kwargs: dict[str, Any],
    args: argparse.Namespace,
) -> float:
    if impl == IMPL_TORCH_COMPILE:
        f = lambda: apply_token_bitmask_inplace_torch_compile(logits, bitmask, **kwargs)
    elif impl == IMPL_TRITON:
        f = lambda: apply_token_bitmask_inplace_triton(logits, bitmask, **kwargs)
    else:
        f = lambda: apply_token_bitmask_inplace_cuda(logits, bitmask, **kwargs)

    f()
    torch.testing.assert_close(logits, logits_expected.to("cuda"))

    torch.cuda.synchronize()
    exec_time = do_bench(f, warmup=args.warmup, rep=args.rep)
    return exec_time * 1000


def bench_single_setup(batch_size: int, masked_cnt: int, args: argparse.Namespace) -> list[float]:
    vocab_size = args.vocab_size
    stride = args.stride
    logits_dtype = getattr(torch, args.logits_dtype)
    logits = torch.randn(batch_size, vocab_size, dtype=logits_dtype, device="cuda")
    if masked_cnt >= vocab_size:
        bool_mask = torch.zeros(batch_size, vocab_size, dtype=torch.bool, device="cuda")
    else:
        bool_mask = torch.ones(batch_size, vocab_size, dtype=torch.bool, device="cuda")
        if masked_cnt > 0:
            masked_positions = torch.stack(
                [torch.randperm(vocab_size, device="cuda")[:masked_cnt] for _ in range(batch_size)]
            )
            bool_mask.scatter_(1, masked_positions, False)
            assert (bool_mask.sum(dim=-1) + masked_cnt == vocab_size).all().item()
    bitmask = _bool_mask_to_bitmask(bool_mask)
    masked_batch_ids = torch.arange(0, batch_size, stride, dtype=torch.int32, device="cuda")
    kwargs = {} if stride == 1 else {"indices": masked_batch_ids}

    logits_copies = [logits.clone() for _ in range(len(args.impl))]
    logits[masked_batch_ids] = torch.masked_fill(
        logits[masked_batch_ids], ~bool_mask[masked_batch_ids], float("-inf")
    )
    return [
        bench_single_impl(impl, logits_copy, bitmask, logits, kwargs, args)
        for impl, logits_copy in zip(args.impl, logits_copies)
    ]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--impl",
        type=str,
        nargs="*",
        choices=ALL_IMPLS,
        default=[IMPL_TORCH_COMPILE, IMPL_TRITON, IMPL_CUDA],
    )
    parser.add_argument("--batch-size", type=int, nargs="*", default=[1, 8, 64, 512, 4096])
    parser.add_argument("--vocab-size", type=int, default=128000)
    parser.add_argument("--masked-cnt", type=int, nargs="*", default=[1, 64000, 127000])
    parser.add_argument("--stride", type=int, default=1)
    parser.add_argument(
        "--logits_dtype", type=str, choices=["float32", "float16", "bfloat16"], default="float32"
    )
    parser.add_argument("--warmup", type=int, default=500)
    parser.add_argument("--rep", type=int, default=2000)
    args = parser.parse_args()

    data_rows = []
    for batch_size, masked_cnt in tqdm(list(product(args.batch_size, args.masked_cnt))):
        all_us = bench_single_setup(batch_size, masked_cnt, args)
        data_rows.append(
            [
                batch_size,
                args.vocab_size,
                masked_cnt,
                f"{all_us[0]:.2f}",
                *[f"{us:.2f} ({all_us[0]/us:>4.2f}x)" for us in all_us[1:]],
            ]
        )

    print(
        tabulate(
            data_rows,
            headers=[
                "Batch\nsize",
                "Vocab\nsize",
                "Masked cnt",
                f"{args.impl[0]}\nBaseline us",
                *[f"{impl} \nus (speedup)" for impl in args.impl[1:]],
            ],
            tablefmt="pipe",
            floatfmt=".2f",
            colalign=["right"] * len(data_rows[0]),
        )
    )



from typing import List, Optional

import torch

try:
    import triton
    import triton.language as tl
except ImportError as err:
    raise ImportError("Triton is not installed") from err


@triton.jit
def apply_token_bitmask_inplace_kernel(
    logits_ptr,
    bitmask_ptr,
    indices_ptr,
    num_rows,
    vocab_size,
    logits_strides,
    bitmask_strides,
    NUM_SMS: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    """Apply a bitmask to logits in-place using Triton. The bitmask is a 01 bitwise compressed tensor,
    where 0 means the token is masked and 1 means the token is not masked. After applying the bitmask,
    the masked logits will be set to -inf.

    Parameters
    ----------
    logits_ptr : tl.tensor
        Pointer to the logits tensor to apply the bitmask to.

    bitmask_ptr : tl.tensor
        Pointer to the bitmask tensor to apply.

    indices_ptr : Optional[tl.tensor]
        Optional pointer to indices tensor specifying which rows to apply the mask to.

    num_rows : int
        Number of rows to process. If indices_ptr is provided, this is the number of unique indices.

    vocab_size : int
        Size of the vocabulary dimension. If the logits does not have a vocab padding, this is the
        same as the logits's second dimension. Otherwise, this is the actual size of the vocabulary.

    logits_strides : int
        Stride between rows in the logits tensor.

    bitmask_strides : int
        Stride between rows in the bitmask tensor.

    NUM_SMS : int
        Number of streaming multiprocessors to use.

    BLOCK_SIZE : int
        Size of processing blocks.
    """

    pid = tl.program_id(0)
    num_blocks = tl.cdiv(vocab_size, BLOCK_SIZE)
    for work_id in tl.range(pid, num_rows * num_blocks, NUM_SMS):
        row_id = work_id // num_blocks
        block_offset = (work_id % num_blocks) * BLOCK_SIZE
        batch_id = row_id if indices_ptr is None else tl.load(indices_ptr + row_id)
        offsets = block_offset + tl.arange(0, BLOCK_SIZE)
        bitmask_offsets = block_offset // 32 + tl.arange(0, BLOCK_SIZE // 32)
        vocab_mask = offsets < vocab_size
        packed_bitmask_mask = bitmask_offsets < bitmask_strides
        packed_bitmask = tl.load(
            bitmask_ptr + batch_id * bitmask_strides + bitmask_offsets, packed_bitmask_mask
        )
        bitmask = ((packed_bitmask[:, None] >> (tl.arange(0, 32)[None, :])) & 1) == 0
        bitmask = bitmask.reshape(BLOCK_SIZE)

        tl.store(
            logits_ptr + batch_id * logits_strides + offsets, -float("inf"), vocab_mask & bitmask
        )


def apply_token_bitmask_inplace_triton(
    logits: torch.Tensor,
    bitmask: torch.Tensor,
    vocab_size: Optional[int] = None,
    indices: Optional[List[int]] = None,
):
    NUM_SMS = torch.cuda.get_device_properties("cuda").multi_processor_count
    BLOCK_SIZE = 4096

    arch = torch.cuda.get_device_properties(0).gcnArchName
    if torch.version.hip is not None and "gfx1" not in arch:
        # For AMD GPUs (non-Navi)
        WARP_SIZE = 64
    else:
        WARP_SIZE = 32

    assert bitmask.dtype == torch.int32, "bitmask must be of type int32"

    detected_vocab_size = min(logits.shape[-1], bitmask.shape[-1] * 32)
    if vocab_size is None:
        vocab_size = detected_vocab_size
    else:
        assert (
            vocab_size <= detected_vocab_size
        ), f"vocab_size {vocab_size} is larger than the detected vocab_size {detected_vocab_size}"

    num_rows = len(indices) if indices is not None else logits.shape[0] if logits.ndim == 2 else 1

    if indices is not None:
        indices_cpu = torch.tensor(indices, dtype=torch.int32)
        indices = indices_cpu.to(device=logits.device, non_blocking=True)

    grid = (NUM_SMS,)

    apply_token_bitmask_inplace_kernel[grid](
        logits,
        bitmask,
        indices,
        num_rows,
        vocab_size,
        logits.stride()[0],
        bitmask.stride()[0],
        NUM_SMS,
        BLOCK_SIZE,
        num_warps=BLOCK_SIZE // WARP_SIZE // (16 // logits.element_size()),
        num_stages=3,
    )



"""Test the basic functionality of GrammarMatcher."""

import math
import random
import sys
from typing import List, Optional, Union

import pytest
import torch
from transformers import AutoTokenizer

import xgrammar as xgr
from xgrammar.testing import (
    _get_masked_tokens_from_bitmask,
    _get_matcher_from_grammar,
    _get_matcher_from_grammar_and_tokenizer_info,
    _is_grammar_accept_string,
)

_is_cuda_available = torch.cuda.is_available()

json_grammar = xgr.Grammar.builtin_json_grammar()


grammar__input__accepted__test_accept_string = [
    ("""root ::= [^a]+""", "bbb", True),
    ("""root ::= [^a]+""", "bba", False),
    ("""root ::= [^a]+""", "©", True),
    ("""root ::= [^a]+""", b"\xe2\xa1\xa1", True),
    ("""root ::= [^a]+""", b"\xe2\xa1\xa1\xa1", False),
    ("""root ::= [^a]+""", b"\xe2\xa1\xe2\xa1", False),
]


@pytest.mark.parametrize("grammar, input, accepted", grammar__input__accepted__test_accept_string)
def test_accept_string(grammar: str, input: Union[str, bytes], accepted: bool):
    matcher = _get_matcher_from_grammar(grammar)
    assert matcher.accept_string(input) == accepted


input_accepted = ['{"name": "John"}', '{ "name" : "John" }']


@pytest.mark.parametrize("input_accepted", input_accepted)
def test_grammar_accept(input_accepted: str):
    assert _is_grammar_accept_string(json_grammar, input_accepted)


input_refused = ('{ name: "John" }', '{ "name": "John" } ')


@pytest.mark.parametrize("input_refused", input_refused)
def test_grammar_refuse(input_refused: str):
    assert not _is_grammar_accept_string(json_grammar, input_refused)


def test_debug_print_internal_state():
    matcher = _get_matcher_from_grammar(json_grammar)
    input_str = '{"name": "John"}'
    for c in input_str:
        assert matcher.accept_string(c)
        internal_state = matcher._debug_print_internal_state()
        assert len(internal_state) > 0


tokenizer_path__input_str__expected_rejected_sizes = [
    (
        "meta-llama/Llama-2-7b-chat-hf",
        '{"id": 1,"name": "Example"}',
        [
            # fmt: off
            31989, 31912, 270, 270, 270, 31973, 31846, 31846, 31948, 31915, 270, 270, 270, 270,
            270, 31973, 31846, 31846, 263, 263, 263, 263, 263, 263, 263, 263, 31974, 31999,
            # fmt: on
        ],
    ),
    (
        # test for llama 3
        "meta-llama/Meta-Llama-3-8B-Instruct",
        '{"id": 1,"name": "Example哈哈"}',
        [
            # fmt: off
            128235, 127497, 4744, 4744, 4744, 127849, 126399, 126399, 126760, 127499, 4744, 4744,
            4744, 4744, 4744, 127849, 126399, 126399, 4694, 4694, 4694, 4694, 4694, 4694, 4694,
            4694, 128066, 128111, 4694, 128066, 128111, 4694, 127873, 128255,
            # fmt: on
        ],
    ),
]


@pytest.mark.hf_token_required
@pytest.mark.parametrize(
    "tokenizer_path, input_str, expected_rejected_sizes",
    tokenizer_path__input_str__expected_rejected_sizes,
)
def test_fill_next_token_bitmask(
    tokenizer_path: str, input_str: str, expected_rejected_sizes: Optional[List[int]]
):
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)

    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    input_bytes = input_str.encode("utf-8")
    rejected_sizes = []

    for i, c in enumerate(input_bytes):
        matcher.fill_next_token_bitmask(token_bitmask)
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            token_bitmask, tokenizer_info.vocab_size
        )
        rejected_sizes.append(len(rejected_token_ids))
        if expected_rejected_sizes is not None:
            assert rejected_sizes[-1] == expected_rejected_sizes[i], (
                rejected_sizes[-1],
                expected_rejected_sizes[i],
            )
        assert matcher.accept_string(bytes([c]))

    matcher.fill_next_token_bitmask(token_bitmask)
    rejected_token_ids = _get_masked_tokens_from_bitmask(token_bitmask, tokenizer_info.vocab_size)
    rejected_sizes.append(len(rejected_token_ids))
    if expected_rejected_sizes is not None:
        assert rejected_sizes[-1] == expected_rejected_sizes[-1]


def test_token_operations():
    """Test accepting token and finding the next token mask."""
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", "\n", " ", '"a":true',
        # fmt: on
    ]
    input_splitted = ["{", '"', "abc", 'b"', ":", "6", ", ", " ", '"a":true', "}"]
    input_ids = [vocab.index(t) for t in input_splitted]

    tokenizer_info = xgr.TokenizerInfo(vocab)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)
    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    expected = [
        ["{"],
        ['"', "}", "\n", " ", '"a":true'],
        ["<s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", " "],
        ["<s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", " "],
        [":", "\n", " ", ':"'],
        ['"', "{", "6", "\n", " "],
        ["}", ", ", "6", "\n", " "],
        [" ", "\n", '"', '"a":true'],
        [" ", "\n", '"', '"a":true'],
        ["}", ", ", "\n", " "],
        ["</s>"],
    ]

    result = []

    for id in input_ids:
        matcher.fill_next_token_bitmask(token_bitmask)
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            token_bitmask, tokenizer_info.vocab_size
        )
        accepted = list(set(range(len(vocab))) - set(rejected_token_ids))
        accepted_tokens = [vocab[i] for i in accepted]
        result.append(accepted_tokens)
        assert id in accepted, vocab[id]
        assert matcher.accept_token(id)

    matcher.fill_next_token_bitmask(token_bitmask)
    rejected_token_ids = _get_masked_tokens_from_bitmask(token_bitmask, tokenizer_info.vocab_size)
    accepted = list(set(range(len(vocab))) - set(rejected_token_ids))
    accepted_tokens = [vocab[i] for i in accepted]
    result.append(accepted_tokens)

    assert result == expected


def test_rollback():
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", "\n", " ", '"a":true',
        # fmt: on
    ]
    input_splitted = ["{", '"', "abc", 'b"', ":", "6", ", ", " ", '"a":true', "}"]
    input_ids = [vocab.index(t) for t in input_splitted]

    tokenizer_info = xgr.TokenizerInfo(vocab)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(
        json_grammar, tokenizer_info, max_rollback_tokens=5
    )

    assert matcher.max_rollback_tokens == -1

    input_ids_splitted = [input_ids[i : i + 2] for i in range(0, len(input_ids), 2)]

    for i_1, i_2 in input_ids_splitted:
        orig_result = []
        token_bitmask1 = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(token_bitmask1)
        orig_result.append(token_bitmask1)
        assert matcher.accept_token(i_1)
        token_bitmask2 = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(token_bitmask2)
        orig_result.append(token_bitmask2)
        assert matcher.accept_token(i_2)

        matcher.rollback(2)
        result_after_rollback = []
        new_token_bitmask1 = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(new_token_bitmask1)
        result_after_rollback.append(new_token_bitmask1)
        assert matcher.accept_token(i_1)
        new_token_bitmask2 = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(new_token_bitmask2)
        result_after_rollback.append(new_token_bitmask2)
        assert matcher.accept_token(i_2)
        for l, r in zip(orig_result, result_after_rollback):
            torch.testing.assert_close(l, r)


def test_graceful_rollback_failure():
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", "6:", ":", "\n", " ", '"a":true',
        # fmt: on
    ]
    input_splitted = ["{", '"', "abc", '"', ":"]
    input_ids = [vocab.index(t) for t in input_splitted]

    tokenizer_info = xgr.TokenizerInfo(vocab)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(
        json_grammar, tokenizer_info, max_rollback_tokens=5
    )

    for i in input_ids:
        assert matcher.accept_token(i)

    assert not matcher.accept_token(vocab.index("6:"))

    # The matching should have accepted char '6' but failed to accept char ':'
    # A graceful revert should then occur, where char '6' is rolled back and
    # the state of the matcher is the same as before the failed call to accept_token

    for i in map(vocab.index, ['"', "abc", '"', " ", "}"]):
        assert matcher.accept_token(i)


def test_reset():
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", "\n", " ", '"a":true',
        # fmt: on
    ]
    input_splitted = ["{", '"', "abc", 'b"', ":", "6", ", ", " ", '"a":true', "}"]
    input_ids = [vocab.index(t) for t in input_splitted]

    tokenizer_info = xgr.TokenizerInfo(vocab)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)

    orig_result = []

    for i in input_ids:
        token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(token_bitmask)
        orig_result.append(token_bitmask)
        assert matcher.accept_token(i)

    matcher.reset()

    result_after_reset = []

    for i in input_ids:
        token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
        matcher.fill_next_token_bitmask(token_bitmask)
        result_after_reset.append(token_bitmask)
        assert matcher.accept_token(i)

    for l, r in zip(orig_result, result_after_reset):
        torch.testing.assert_close(l, r)


def test_termination():
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", " }", ", ", "6", ":", "\n", " ", '"a"', ':true',
        # fmt: on
    ]
    input_splitted = ["{", '"', "abc", 'b"', ":", "6", ", ", " ", '"a"', ":true", " }", "</s>"]
    input_ids = [vocab.index(t) for t in input_splitted]
    tokenizer_info = xgr.TokenizerInfo(vocab)

    matcher = _get_matcher_from_grammar_and_tokenizer_info(
        json_grammar, tokenizer_info, max_rollback_tokens=5
    )
    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    for i in input_ids:
        matcher.fill_next_token_bitmask(token_bitmask)
        assert matcher.accept_token(i)

    assert matcher.is_terminated()

    assert matcher.accept_token(0) is False

    with pytest.raises(RuntimeError):
        matcher.fill_next_token_bitmask(token_bitmask)

    matcher.rollback(2)

    assert not matcher.is_terminated()
    assert matcher.accept_token(input_ids[-2])


def test_get_jump_forward_string():
    grammar_ebnf = r"""root ::= "abb" | "abbd" | other_rule
other_rule ::= "a" sub_rule "b"
sub_rule ::= "b"
"""
    grammar = xgr.Grammar.from_ebnf(grammar_ebnf)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(grammar)
    assert matcher.accept_string("a")
    assert matcher.find_jump_forward_string() == "bb"


def test_vocab_size():
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "abc", 'b"', '"', ':"', "{", "}", ", ", "6", ":", "\n", " ", '"a":true',
        # fmt: on
    ]
    tokenizer_info = xgr.TokenizerInfo(vocab, vocab_size=64)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)

    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
    matcher.fill_next_token_bitmask(token_bitmask)
    assert token_bitmask.shape == (1, 2)

    rejected_tokens = _get_masked_tokens_from_bitmask(token_bitmask, tokenizer_info.vocab_size)
    assert rejected_tokens == [i for i in range(64) if i != 7]


tokenizer_path_override_stop_tokens = [
    ("meta-llama/Llama-2-7b-chat-hf", [2]),
    ("meta-llama/Meta-Llama-3-8B-Instruct", [128001, 128009]),
    ("deepseek-ai/DeepSeek-Coder-V2-Lite-Instruct", [100001]),
]


@pytest.mark.hf_token_required
@pytest.mark.parametrize(
    "tokenizer_path, override_stop_tokens", tokenizer_path_override_stop_tokens
)
def test_override_stop_tokens(tokenizer_path: str, override_stop_tokens: List[int]):
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True, trust_remote_code=True)
    tokenizer_info_1 = xgr.TokenizerInfo.from_huggingface(
        tokenizer, stop_token_ids=override_stop_tokens
    )
    matcher_1 = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info_1)
    assert tokenizer_info_1.stop_token_ids == override_stop_tokens
    assert matcher_1.stop_token_ids == override_stop_tokens

    tokenizer_info_2 = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matcher_2 = _get_matcher_from_grammar_and_tokenizer_info(
        json_grammar, tokenizer_info_2, override_stop_tokens=override_stop_tokens
    )
    assert matcher_2.stop_token_ids == override_stop_tokens


@pytest.mark.hf_token_required
def test_fill_next_token_bitmask_errors():
    # llama 3.1 8b
    tokenizer = AutoTokenizer.from_pretrained(
        "meta-llama/Meta-Llama-3-8B-Instruct", use_fast=True, trust_remote_code=True
    )
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matcher = _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)

    bitmask1 = torch.zeros(1, math.ceil(tokenizer_info.vocab_size / 32) - 1, dtype=torch.int32)
    with pytest.raises(RuntimeError):
        matcher.fill_next_token_bitmask(bitmask1)

    bitmask2 = torch.zeros(1, math.ceil(tokenizer_info.vocab_size / 32), dtype=torch.int32)
    with pytest.raises(RuntimeError):
        matcher.fill_next_token_bitmask(bitmask2, index=1)

    bitmask3 = torch.zeros(1, math.ceil(tokenizer_info.vocab_size / 32), dtype=torch.float32)
    with pytest.raises(RuntimeError):
        matcher.fill_next_token_bitmask(bitmask3)

    if _is_cuda_available:
        bitmask3 = torch.zeros(1, math.ceil(tokenizer_info.vocab_size / 32), 1, dtype=torch.int32)
        with pytest.raises(RuntimeError):
            matcher.fill_next_token_bitmask(bitmask3)

    bitmask_correct = torch.zeros(1, math.ceil(tokenizer_info.vocab_size / 32), dtype=torch.int32)
    matcher.fill_next_token_bitmask(bitmask_correct)


test_batch_accept_string_grammars_inputs_expecteds = [
    (['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'], ["a", b"123", "ab"], [True, True, True]),
    (
        ['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'],
        ["b", "123a", "d"],
        [False, False, False],
    ),
    (
        ['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'],
        ["a", b"123a", b"ab"],
        [True, False, True],
    ),
    (['root ::= "a"'], ["a"], [True]),
    (['root ::= "a"'], ["b"], [False]),
    (
        ['root ::= "你好"', 'root ::= "こんにちは"', 'root ::= "안녕하세요"'],
        ["你好", "こんにちは", "안녕하세요"],
        [True, True, True],
    ),
]


@pytest.mark.parametrize(
    "grammars, inputs, expecteds", test_batch_accept_string_grammars_inputs_expecteds
)
def test_batch_accept_string(
    grammars: List[str], inputs: List[Union[str, bytes]], expecteds: List[bool]
):
    matchers = [_get_matcher_from_grammar(grammar) for grammar in grammars]
    results = xgr.BatchGrammarMatcher.batch_accept_string(matchers, inputs)
    assert results == expecteds


test_batch_accept_token_grammars_inputs_expecteds = [
    (['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'], [2, 5, 2], [True, True, True]),
    (['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'], [3, 2, 4], [False, False, False]),
    (['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"'], [2, 8, 9], [True, False, True]),
    (['root ::= "a"'], [2], [True]),
    (['root ::= "a"'], [3], [False]),
]


@pytest.mark.parametrize(
    "grammars, inputs, expecteds", test_batch_accept_token_grammars_inputs_expecteds
)
def test_batch_accept_token(grammars: List[str], inputs: List[int], expecteds: List[bool]):
    vocab = [
        # fmt: off
        "<s>", "</s>", "a", "b", "c", "1", "2", "3", "123a", "ab",
        # fmt: on
    ]
    tokenizer_info = xgr.TokenizerInfo(vocab)

    matchers = [
        _get_matcher_from_grammar_and_tokenizer_info(xgr.Grammar.from_ebnf(grammar), tokenizer_info)
        for grammar in grammars
    ]
    results = xgr.BatchGrammarMatcher.batch_accept_token(matchers, inputs)
    assert results == expecteds


def test_batch_fill_next_token_bitmask():
    grammars = ['root ::= "a"', "root ::= [0-9]+", 'root ::= "ab"', "root ::= [a-z0-9]+"]
    vocab = [
        # fmt: off
        "ab", "</s>", "a", "b", "c", "1", "2", "3", "123a"
        # fmt: on
    ]
    tokenizer_info = xgr.TokenizerInfo(vocab)

    matchers = [
        _get_matcher_from_grammar_and_tokenizer_info(xgr.Grammar.from_ebnf(grammar), tokenizer_info)
        for grammar in grammars
    ]

    batch_size = len(matchers)
    token_bitmask = xgr.allocate_token_bitmask(batch_size, tokenizer_info.vocab_size)

    input_str = ["a", "1", "a", "123a"]

    expected_accepted_tokens = [
        [[2], [5, 6, 7], [0, 2], [0, 2, 3, 4, 5, 6, 7, 8]],
        [[1], [1, 5, 6, 7], [3], [0, 1, 2, 3, 4, 5, 6, 7, 8]],
    ]

    batch_grammar_matcher = xgr.BatchGrammarMatcher(2)
    batch_grammar_matcher.batch_fill_next_token_bitmask(matchers, token_bitmask)

    for i in range(batch_size):
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            token_bitmask[i : i + 1], tokenizer_info.vocab_size
        )
        accepted = list(set(range(len(vocab))) - set(rejected_token_ids))
        accepted.sort()
        assert accepted == expected_accepted_tokens[0][i]

    assert xgr.BatchGrammarMatcher.batch_accept_string(matchers, input_str) == [
        True,
        True,
        True,
        True,
    ]

    batch_grammar_matcher.batch_fill_next_token_bitmask(matchers, token_bitmask)

    for i in range(batch_size):
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            token_bitmask[i : i + 1], tokenizer_info.vocab_size
        )
        accepted = list(set(range(len(vocab))) - set(rejected_token_ids))
        accepted.sort()
        assert accepted == expected_accepted_tokens[1][i]


@pytest.mark.hf_token_required
def test_batch_fill_next_token_bitmask_pressure():
    tokenizer_path = "meta-llama/Llama-2-7b-chat-hf"
    input_str = '{"id": 1,"name": "Example"}'
    rejected_token_size = [
        # fmt: off
            31989, 31912, 270, 270, 270, 31973, 31846, 31846, 31948, 31915, 270, 270, 270, 270,
            270, 31973, 31846, 31846, 263, 263, 263, 263, 263, 263, 263, 263, 31974, 31999,
        # fmt: on
    ]
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matchers = [
        _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)
        for _ in range(len(input_str) + 1)
    ]
    input_strs = [input_str[:i] for i in range(len(input_str))] + [input_str]
    xgr.BatchGrammarMatcher.batch_accept_string(matchers, input_strs)

    bitmask_2d = xgr.allocate_token_bitmask(len(matchers), tokenizer_info.vocab_size)
    batch_grammar_matcher = xgr.BatchGrammarMatcher(2)
    batch_grammar_matcher.batch_fill_next_token_bitmask(matchers, bitmask_2d)
    for i in range(len(matchers)):
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            bitmask_2d[i], tokenizer_info.vocab_size
        )
        assert len(rejected_token_ids) == rejected_token_size[i], (
            i,
            len(rejected_token_ids),
            rejected_token_size[i],
        )


@pytest.mark.hf_token_required
def test_batch_fill_next_token_bitmask_pressure_single_thread():
    tokenizer_path = "meta-llama/Llama-2-7b-chat-hf"
    input_str = '{"id": 1,"name": "Example"}'
    rejected_token_size = [
        # fmt: off
            31989, 31912, 270, 270, 270, 31973, 31846, 31846, 31948, 31915, 270, 270, 270, 270,
            270, 31973, 31846, 31846, 263, 263, 263, 263, 263, 263, 263, 263, 31974, 31999,
        # fmt: on
    ]
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matchers = [
        _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)
        for _ in range(len(input_str) + 1)
    ]
    input_strs = [input_str[:i] for i in range(len(input_str))] + [input_str]
    xgr.BatchGrammarMatcher.batch_accept_string(matchers, input_strs)

    bitmask_2d = xgr.allocate_token_bitmask(len(matchers), tokenizer_info.vocab_size)
    batch_grammar_matcher = xgr.BatchGrammarMatcher(1)
    batch_grammar_matcher.batch_fill_next_token_bitmask(matchers, bitmask_2d)
    for i in range(len(matchers)):
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            bitmask_2d[i], tokenizer_info.vocab_size
        )
        assert len(rejected_token_ids) == rejected_token_size[i], (
            i,
            len(rejected_token_ids),
            rejected_token_size[i],
        )


@pytest.mark.hf_token_required
def test_batch_fill_next_token_bitmask_pressure_shuffled():
    tokenizer_path = "meta-llama/Llama-2-7b-chat-hf"
    input_str = '{"id": 1,"name": "Example"}'
    rejected_token_size = [
        # fmt: off
            31989, 31912, 270, 270, 270, 31973, 31846, 31846, 31948, 31915, 270, 270, 270, 270,
            270, 31973, 31846, 31846, 263, 263, 263, 263, 263, 263, 263, 263, 31974, 31999,
        # fmt: on
    ]
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True, trust_remote_code=True)
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    matchers = [
        _get_matcher_from_grammar_and_tokenizer_info(json_grammar, tokenizer_info)
        for _ in range(len(input_str) + 1)
    ]
    input_strs = [input_str[:i] for i in range(len(input_str))] + [input_str]
    xgr.BatchGrammarMatcher.batch_accept_string(matchers, input_strs)

    shuffled_indices = list(range(len(matchers)))
    random.shuffle(shuffled_indices)
    bitmask_2d = xgr.allocate_token_bitmask(len(matchers), tokenizer_info.vocab_size)
    batch_grammar_matcher = xgr.BatchGrammarMatcher()
    batch_grammar_matcher.batch_fill_next_token_bitmask(matchers, bitmask_2d, shuffled_indices)
    for i in range(len(matchers)):
        rejected_token_ids = _get_masked_tokens_from_bitmask(
            bitmask_2d[shuffled_indices[i]], tokenizer_info.vocab_size
        )
        assert len(rejected_token_ids) == rejected_token_size[i], (
            i,
            len(rejected_token_ids),
            rejected_token_size[i],
        )


if __name__ == "__main__":
    pytest.main(sys.argv)



# -*- coding: utf-8 -*-
import json
import sys
from typing import Any, List, Tuple

import pytest
from pydantic import BaseModel, RootModel
from transformers import AutoTokenizer  # type: ignore

import xgrammar as xgr
from xgrammar.testing import _is_grammar_accept_string


def construct_grammar():
    """Construct a Grammar object for testing."""
    return xgr.Grammar.from_ebnf(
        """rule1 ::= ([^0-9] rule1) | ""
root_rule ::= rule1 "a"
""",
        root_rule_name="root_rule",
    )


def construct_tokenizer_info():
    """Construct a TokenizerInfo object for testing."""
    return xgr.TokenizerInfo(
        ["1", "212", "a", "A", "b", "一", "-", "aBc", "abc"],
        vocab_type=xgr.VocabType.BYTE_FALLBACK,
        vocab_size=10,
        stop_token_ids=[0, 1],
        add_prefix_space=True,
    )


def construct_compiled_grammar():
    """Construct a CompiledGrammar object for testing."""
    tokenizer_info = construct_tokenizer_info()
    grammar = construct_grammar()
    grammar_compiler = xgr.GrammarCompiler(tokenizer_info)
    return grammar_compiler.compile_grammar(grammar), tokenizer_info


def test_get_serialization_version():
    """Test the version of the serialized JSON string."""
    assert xgr.get_serialization_version() == "v7"


def test_serialize_grammar():
    """Test Grammar serialization produces expected JSON string."""
    grammar = construct_grammar()
    serialized = grammar.serialize_json()
    expected_json = {
        "rules": [["rule1", 4, -1, False], ["root_rule", 8, -1, False]],
        "grammar_expr_data": [0, 5, 8, 12, 14, 18, 21, 24, 28],
        "grammar_expr_indptr": [
            # fmt: off
            1,3,1,48,57,4,1,0,5,2,0,1,3,0,6,2,3,2,4,1,0,0,1,97,5,2,5,6,6,1,7
            # fmt: on
        ],
        "root_rule_id": 1,
        "complete_fsm": None,
        "per_rule_fsms": [],
        "allow_empty_rule_ids": [],
        "optimized": False,
        "__VERSION__": "v7",
    }
    # The fsms are the same one, but the start state and end states are different.
    assert json.loads(serialized) == expected_json


def test_serialize_grammar_exception():
    """Test Grammar serialization produces expected JSON string."""
    expected_json = {
        "rules": [["rule1", 4, 9, True], ["root_rule", 8, -1, False]],
        "grammar_expr_data": [0, 2, 7, 10, 14, 18, 21, 24, 28, 31],
        "grammar_expr_indptr": [
            # fmt: off
            3,0,1,3,1,48,57,4,1,0,5,2,1,2,6,2,0,3,4,1,0,0,1,97,5,2,5,6,6,1,7,5,1,6
            # fmt: on
        ],
        "root_rule_id": 1,
        "allow_empty_rule_ids": [],
        "complete_fsm": None,
        "per_rule_fsms": [],
        "__VERSION__": "v7",
    }

    expected_json["__VERSION__"] = "v1"  # Change version to trigger error
    with pytest.raises(xgr.DeserializeVersionError):
        xgr.Grammar.deserialize_json(json.dumps(expected_json))

    expected_json["__VERSION__"] = "v7"
    expected_json.pop("rules")  # Remove required field to trigger error
    with pytest.raises(xgr.DeserializeFormatError):
        xgr.Grammar.deserialize_json(json.dumps(expected_json))

    with pytest.raises(xgr.InvalidJSONError):
        xgr.Grammar.deserialize_json("not a valid json string")


def test_serialize_grammar_roundtrip():
    """Test Grammar serialization and deserialization roundtrip."""
    original_grammar = construct_grammar()
    serialized = original_grammar.serialize_json()
    recovered_grammar = xgr.Grammar.deserialize_json(serialized)
    serialized_new = recovered_grammar.serialize_json()
    assert serialized == serialized_new


def test_serialize_grammar_functional():
    """Test that deserialized Grammar object functions correctly."""
    original_grammar = construct_grammar()
    serialized = original_grammar.serialize_json()
    recovered_grammar = xgr.Grammar.deserialize_json(serialized)

    # Test functional equivalence by checking string representation
    assert str(original_grammar) == str(recovered_grammar)

    # Test with GrammarMatcher functionality
    tokenizer_info = construct_tokenizer_info()
    compiler = xgr.GrammarCompiler(tokenizer_info)

    compiled_original = compiler.compile_grammar(original_grammar)
    compiled_recovered = compiler.compile_grammar(recovered_grammar)

    matcher_original = xgr.GrammarMatcher(compiled_original)
    matcher_recovered = xgr.GrammarMatcher(compiled_recovered)

    # Test that both matchers accept the same input
    test_input = "aaa"
    assert matcher_original.accept_string(test_input) == matcher_recovered.accept_string(test_input)


def test_serialize_tokenizer_info():
    """Test TokenizerInfo serialization produces expected JSON string."""
    tokenizer_info = construct_tokenizer_info()
    serialized = tokenizer_info.serialize_json()
    expected_json = (
        '{"vocab_type":1,"vocab_size":10,"add_prefix_space":true,'
        '"stop_token_ids":[0,1],"special_token_ids":[9],'
        '"decoded_vocab":["1","212","a","A","b","\\u00e4\\u00b8\\u0080","-","aBc","abc"],'
        '"sorted_decoded_vocab":[[6,"-"],[3,"A"],[2,"a"],[7,"aBc"],[8,"abc"],[4,"b"],[5,"\\u00e4\\u00b8\\u0080"]],'
        '"trie_subtree_nodes_range":[1,2,5,4,5,6,7],'
        '"__VERSION__":"v7"}'
    )
    assert json.loads(serialized) == json.loads(expected_json)


def test_serialize_tokenizer_info_roundtrip():
    """Test TokenizerInfo serialization and deserialization roundtrip."""
    original_tokenizer_info = construct_tokenizer_info()
    serialized = original_tokenizer_info.serialize_json()
    recovered_tokenizer_info = xgr.TokenizerInfo.deserialize_json(serialized)
    serialized_new = recovered_tokenizer_info.serialize_json()
    assert serialized == serialized_new


def test_serialize_tokenizer_info_functional():
    """Test that deserialized TokenizerInfo object functions correctly."""
    original_tokenizer_info = construct_tokenizer_info()
    serialized = original_tokenizer_info.serialize_json()
    recovered_tokenizer_info = xgr.TokenizerInfo.deserialize_json(serialized)

    # Test property equivalence
    assert original_tokenizer_info.vocab_type == recovered_tokenizer_info.vocab_type
    assert original_tokenizer_info.vocab_size == recovered_tokenizer_info.vocab_size
    assert original_tokenizer_info.add_prefix_space == recovered_tokenizer_info.add_prefix_space
    assert original_tokenizer_info.stop_token_ids == recovered_tokenizer_info.stop_token_ids
    assert original_tokenizer_info.special_token_ids == recovered_tokenizer_info.special_token_ids
    assert original_tokenizer_info.decoded_vocab == recovered_tokenizer_info.decoded_vocab

    # Test functional equivalence with GrammarCompiler
    grammar = construct_grammar()

    compiler_original = xgr.GrammarCompiler(original_tokenizer_info)
    compiler_recovered = xgr.GrammarCompiler(recovered_tokenizer_info)

    compiled_original = compiler_original.compile_grammar(grammar)
    compiled_recovered = compiler_recovered.compile_grammar(grammar)

    # Both should produce functional matchers
    matcher_original = xgr.GrammarMatcher(compiled_original)
    matcher_recovered = xgr.GrammarMatcher(compiled_recovered)

    test_input = "aaa"
    assert matcher_original.accept_string(test_input) == matcher_recovered.accept_string(test_input)


def test_serialize_compiled_grammar():
    """Test CompiledGrammar serialization produces expected JSON string. We verify the adaptive
    token mask part separately.
    """
    compiled_grammar, tokenizer_info = construct_compiled_grammar()
    serialized = compiled_grammar.serialize_json()

    expected_json = {
        "grammar": {
            "rules": [["rule1", 4, 9, True], ["root_rule", 8, -1, False]],
            "grammar_expr_data": [0, 2, 7, 10, 14, 18, 21, 24, 28, 31],
            "grammar_expr_indptr": [
                # fmt: off
                3,0,1,3,1,48,57,4,1,0,5,2,1,2,6,2,0,3,4,1,0,0,1,97,5,2,5,6,6,1,7,5,1,6
                # fmt: on
            ],
            "root_rule_id": 1,
            "allow_empty_rule_ids": [0],
            # fmt: off
            "complete_fsm": {
                'data_': [[0, 47, 3], [58, 127, 3], [192, 223, 1], [224, 239, 4], [240, 247, 5], [128, 191, 3], [-2, 0, 2], [128, 191, 1], [128, 191, 4], [-2, 0, 8], [97, 97, 6]],
                'indptr_': [0, 5, 6, 6, 7, 8, 9, 9, 10, 11]
                },
            "per_rule_fsms": [
                [{'data_': [[0, 47, 3], [58, 127, 3], [192, 223, 1], [224, 239, 4], [240, 247, 5], [128, 191, 3], [-2, 0, 2], [128, 191, 1], [128, 191, 4], [-2, 0, 8], [97, 97, 6]],
                'indptr_': [0, 5, 6, 6, 7, 8, 9, 9, 10, 11]}, 0, [0, 2], False],
                [{'data_': [[0, 47, 3], [58, 127, 3], [192, 223, 1], [224, 239, 4], [240, 247, 5], [128, 191, 3], [-2, 0, 2], [128, 191, 1], [128, 191, 4], [-2, 0, 8], [97, 97, 6]],
                'indptr_': [0, 5, 6, 6, 7, 8, 9, 9, 10, 11]}, 7, [6], False]],
            # fmt: on
            "optimized": True,
        },
        "tokenizer_metadata": {
            "vocab_type": 1,
            "vocab_size": 10,
            "add_prefix_space": True,
            "stop_token_ids": [0, 1],
        },
        "__VERSION__": "v7",
    }

    class AdaptiveTokenMask(BaseModel):
        store_type: int
        accepted_indices: List[int]
        rejected_indices: List[int]
        accepted_bitset: Any
        uncertain_indices: List[int]

    class AdaptiveTokenMaskCache(RootModel):
        root: List[Tuple[List[int], AdaptiveTokenMask]]

    recovered_obj = json.loads(serialized)
    adaptive_token_mask_cache = recovered_obj.pop("adaptive_token_mask_cache", None)
    assert recovered_obj == expected_json
    AdaptiveTokenMaskCache.model_validate(adaptive_token_mask_cache)


def test_serialize_compiled_grammar_roundtrip():
    """Test CompiledGrammar serialization and deserialization roundtrip."""
    original_compiled_grammar, tokenizer_info = construct_compiled_grammar()
    serialized = original_compiled_grammar.serialize_json()
    recovered_compiled_grammar = xgr.CompiledGrammar.deserialize_json(serialized, tokenizer_info)
    serialized_new = recovered_compiled_grammar.serialize_json()
    assert serialized == serialized_new


def test_serialize_compiled_grammar_functional():
    """Test that deserialized CompiledGrammar object functions correctly."""
    original_compiled_grammar, tokenizer_info = construct_compiled_grammar()
    serialized = original_compiled_grammar.serialize_json()
    recovered_compiled_grammar = xgr.CompiledGrammar.deserialize_json(serialized, tokenizer_info)

    # Test that both create functional matchers
    matcher_original = xgr.GrammarMatcher(original_compiled_grammar)
    matcher_recovered = xgr.GrammarMatcher(recovered_compiled_grammar)

    # Test token mask generation
    token_bitmask_original = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)
    token_bitmask_recovered = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    # Both should generate the same masks
    assert matcher_original.fill_next_token_bitmask(
        token_bitmask_original
    ) == matcher_recovered.fill_next_token_bitmask(token_bitmask_recovered)

    # Import torch for tensor comparison
    import torch

    torch.testing.assert_close(token_bitmask_original, token_bitmask_recovered)

    # Test input acceptance
    test_input = "aaa"
    assert matcher_original.accept_string(test_input) == matcher_recovered.accept_string(test_input)
    assert matcher_original.is_terminated() == matcher_recovered.is_terminated()


@pytest.mark.hf_token_required
def test_serialize_compiled_grammar_with_hf_tokenizer():
    """Test CompiledGrammar serialization with a real HuggingFace tokenizer."""
    tokenizer = AutoTokenizer.from_pretrained(
        "meta-llama/Llama-3.1-8B-Instruct", use_fast=True, trust_remote_code=True
    )
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(tokenizer)
    grammar_compiler = xgr.GrammarCompiler(tokenizer_info)

    # Test with JSON schema
    class TestModel(BaseModel):
        name: str
        age: int

    # Compile grammar
    compiled_grammar = grammar_compiler.compile_json_schema(TestModel)

    # Serialize and deserialize
    tokenizer_info_json = tokenizer_info.serialize_json()
    tokenizer_info_recovered = xgr.TokenizerInfo.deserialize_json(tokenizer_info_json)
    serialized = compiled_grammar.serialize_json()
    recovered_compiled_grammar = xgr.CompiledGrammar.deserialize_json(
        serialized, tokenizer_info_recovered
    )

    # Test functional equivalence
    test_json = '{"name": "John", "age": 30}'
    token_ids = tokenizer.encode(test_json)[1:]  # skip the initial BOS token
    matcher = xgr.GrammarMatcher(recovered_compiled_grammar)
    bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    for token_id in token_ids:
        matcher.fill_next_token_bitmask(bitmask)
        masked_token_ids = xgr.testing._get_masked_tokens_from_bitmask(
            bitmask, tokenizer_info.vocab_size
        )
        assert token_id not in masked_token_ids
        assert matcher.accept_token(token_id)

    assert matcher.accept_token(tokenizer.eos_token_id)
    assert matcher.is_terminated()


def test_serialize_grammar_utf8():
    """Test Grammar serialization with UTF-8 characters."""
    grammar = xgr.Grammar.from_ebnf('root ::= "こんにちは" | "😊" | "你好" | "hello" | "\\n"')
    serialized = grammar.serialize_json()
    recovered_grammar = xgr.Grammar.deserialize_json(serialized)
    assert _is_grammar_accept_string(recovered_grammar, "こんにちは")
    assert _is_grammar_accept_string(recovered_grammar, "😊")
    assert _is_grammar_accept_string(recovered_grammar, "你好")
    assert _is_grammar_accept_string(recovered_grammar, "hello")
    assert _is_grammar_accept_string(recovered_grammar, "\n")


if __name__ == "__main__":
    pytest.main(sys.argv)



import sys

import pytest

import xgrammar as xgr
from xgrammar.testing import _get_masked_tokens_from_bitmask, _is_grammar_accept_string


def test_simple():
    grammar_str = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2))
rule1 ::= "abcd"
rule2 ::= "efg"
"""

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "tag1abcd")
    assert _is_grammar_accept_string(grammar, "tag1abcdtag2efg")
    assert _is_grammar_accept_string(grammar, "tag1abcdqqqqtag2efg")
    assert not _is_grammar_accept_string(grammar, "tag1abc")
    assert not _is_grammar_accept_string(grammar, "tag1abce")
    assert not _is_grammar_accept_string(grammar, "ttag1abd")


def test_complex_rule():
    grammar_str = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2))
rule1 ::= "abcd" [p]*
rule2 ::= "efg" [t]*
"""

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "tag1abcd")
    assert _is_grammar_accept_string(grammar, "tag1abcdppppptag2efg")
    assert _is_grammar_accept_string(grammar, "tag2efgtttttag1abc")
    assert not _is_grammar_accept_string(grammar, "tag1efg")


def test_no_loop_after_dispatch():
    grammar_str = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), loop_after_dispatch=false)
rule1 ::= "abcd" [p]*
rule2 ::= "efg" [t]*
"""

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "tag1abcd")
    assert _is_grammar_accept_string(grammar, "tag2efgttt")
    assert not _is_grammar_accept_string(grammar, "tag1abcdppppptag2")
    assert not _is_grammar_accept_string(grammar, "tag2efgtag1")


def test_stop_str():
    grammar_str = """root ::= root1 "w"
root1 ::= TagDispatch(
  ("tag1", rule1),
  ("tag2", rule2),
  stop_eos=false,
  stop_str=("tag3", "ll")
)
rule1 ::= "abcd" [p]*
rule2 ::= "efg" [t]*
"""

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "tag1abcdllw", debug_print=True)
    assert _is_grammar_accept_string(grammar, "tag1abcdtag3w")
    assert _is_grammar_accept_string(grammar, "tag1abcdqqqtag2efgtag3w")
    assert _is_grammar_accept_string(grammar, "tag1abcd", require_termination=False)
    assert _is_grammar_accept_string(grammar, "tag2efgttt", require_termination=False)
    assert not _is_grammar_accept_string(grammar, "tag1abcd")
    assert not _is_grammar_accept_string(grammar, "tag2efgttt")
    assert not _is_grammar_accept_string(grammar, "tag1abce")
    assert not _is_grammar_accept_string(grammar, "tag1abcdlltag3w", require_termination=False)


def test_stop_str_no_loop():
    grammar_str = """root ::= root1 "w"
root1 ::= TagDispatch(
  ("tag1", rule1),
  ("tag2", rule2),
  stop_eos=false,
  stop_str=("tag3", "ll"),
  loop_after_dispatch=false
)
rule1 ::= "abcd" [p]*
rule2 ::= "efg" [t]*
"""

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    assert _is_grammar_accept_string(grammar, "tag1abcdllw")
    assert _is_grammar_accept_string(grammar, "tag1abcdtag3w")
    assert _is_grammar_accept_string(grammar, "tag1abcd", require_termination=False)
    assert _is_grammar_accept_string(grammar, "tag2efgttt", require_termination=False)
    assert not _is_grammar_accept_string(grammar, "tag1abcdqqqtag2efgtag3w")
    assert not _is_grammar_accept_string(grammar, "tag1abcd")
    assert not _is_grammar_accept_string(grammar, "tag2efgttt")
    assert not _is_grammar_accept_string(grammar, "tag1abce")
    assert not _is_grammar_accept_string(grammar, "tag1abcdlltag3w", require_termination=False)


def test_tag_dispatch_mask_generation_correctness():
    grammar_str = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2))
rule1 ::= "abc"
rule2 ::= "dg"
"""
    tokens = [
        # fmt: off
        "a", "b", "c", "d", "g", "t", "1", "2", "1a", "2d", "2a", "2dgt",
        "2dgtag1a", "2dgtag1b", "tag1a", "tag1b", "c哈哈t", "q", "abcdef"
        # fmt: on
    ]
    input_str = "tag1abcqqtag2dgq"
    expected_accepted_tokens = [
        # fmt: off
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'abcdef'],
        ['b'],
        ['c哈哈t', 'c'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['d'],
        ['g'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef'],
        ['a', 'b', 'c', 'd', 'g', 't', '1', '2', '1a', '2d', '2a', '2dgt', '2dgtag1a', 'tag1a', 'c哈哈t', 'q', 'abcdef']
        # fmt: on
    ]

    grammar = xgr.Grammar.from_ebnf(grammar_str)
    tokenizer_info = xgr.TokenizerInfo(tokens)
    compiler = xgr.GrammarCompiler(tokenizer_info, max_threads=1)
    compiled_grammar = compiler.compile_grammar(grammar)
    matcher = xgr.GrammarMatcher(compiled_grammar, terminate_without_stop_token=True)
    mask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    # pad a dummy char to check the final bitmask after accepting the input string
    for i, c in enumerate(input_str + "0"):
        matcher.fill_next_token_bitmask(mask)
        rejected_indices = _get_masked_tokens_from_bitmask(mask, tokenizer_info.vocab_size)
        accepted_indices = list(set(range(tokenizer_info.vocab_size)) - set(rejected_indices))
        accepted_tokens = [tokens[id] for id in accepted_indices]
        if i < len(input_str):
            assert matcher.accept_string(c)
        assert accepted_tokens == expected_accepted_tokens[i]


def test_regression_multiple_tag_dispatch():
    grammar_str = """root ::= root1 "w"
root1 ::= TagDispatch(
  ("tag1", rule1),
  ("tag2", rule2),
  stop_eos=true,
  stop_str=(),
  loop_after_dispatch=false
)
rule1 ::= TagDispatch(
  ("tag1", rule2),
  ("tag2", rule3),
  stop_eos=false,
  stop_str=("tag3", "ll"),
  loop_after_dispatch=true
)
rule2 ::= "efg" [t]*
rule3 ::= "abcd" [p]*
"""
    assert _is_grammar_accept_string(grammar_str, "tag1tag1efgllw")
    assert _is_grammar_accept_string(grammar_str, "tag1tag2abcdtag3w")
    assert not _is_grammar_accept_string(grammar_str, "tag1Ktag2abcdtag3tag1")
    assert _is_grammar_accept_string(grammar_str, "tag1tag3w")
    assert not _is_grammar_accept_string(grammar_str, "tag1tag3tag2abcdll")


if __name__ == "__main__":
    pytest.main(sys.argv)
